from __future__ import annotations
from typing import TYPE_CHECKING

import rich.repr

from .message import Message


if TYPE_CHECKING:
    from .message_pump import MessagePump
    from .widget import Widget


@rich.repr.auto
class UpdateMessage(Message):
    def __init__(
        self,
        sender: MessagePump,
        widget: Widget,
        offset_x: int = 0,
        offset_y: int = 0,
        reflow: bool = False,
    ):
        super().__init__(sender)
        self.widget = widget
        self.offset_x = offset_x
        self.offset_y = offset_y
        self.reflow = reflow

    def __rich_repr__(self) -> rich.repr.RichReprResult:
        yield self.sender
        yield "widget"
        yield "offset_x", self.offset_x, 0
        yield "offset_y", self.offset_y, 0
        yield "reflow", self.reflow, False

    def can_replace(self, message: Message) -> bool:
        return isinstance(message, UpdateMessage) and message.sender == self.sender


@rich.repr.auto
class LayoutMessage(Message):
    def can_replace(self, message: Message) -> bool:
        return isinstance(message, LayoutMessage) and message.sender == self.sender



from __future__ import annotations

import weakref
from asyncio import CancelledError, Event, TimeoutError, wait_for
from time import monotonic
from typing import Awaitable, Callable

from rich.repr import RichReprResult, rich_repr

from . import events
from ._types import MessageTarget

TimerCallback = Callable[[], Awaitable[None]]


class EventTargetGone(Exception):
    pass


@rich_repr
class Timer:
    _timer_count: int = 1

    def __init__(
        self,
        event_target: MessageTarget,
        interval: float,
        sender: MessageTarget,
        *,
        name: str | None = None,
        callback: TimerCallback | None = None,
        repeat: int = None,
        skip: bool = False,
    ) -> None:
        self._target_repr = repr(event_target)
        self._target = weakref.ref(event_target)
        self._interval = interval
        self.sender = sender
        self.name = f"Timer#{self._timer_count}" if name is None else name
        self._timer_count += 1
        self._callback = callback
        self._repeat = repeat
        self._skip = skip
        self._stop_event = Event()
        self._active = Event()
        self._active.set()

    def __rich_repr__(self) -> RichReprResult:
        yield self._interval
        yield "name", self.name
        yield "repeat", self._repeat, None

    @property
    def target(self) -> MessageTarget:
        target = self._target()
        if target is None:
            raise EventTargetGone()
        return target

    def stop(self) -> None:
        self._active.set()
        self._stop_event.set()

    def pause(self) -> None:
        self._active.clear()

    def resume(self) -> None:
        self._active.set()

    async def run(self) -> None:
        count = 0
        _repeat = self._repeat
        _interval = self._interval
        _wait = self._stop_event.wait
        _wait_active = self._active.wait
        start = monotonic()
        try:
            while _repeat is None or count <= _repeat:
                next_timer = start + ((count + 1) * _interval)
                if self._skip and next_timer < monotonic():
                    count += 1
                    continue
                try:
                    if await wait_for(_wait(), max(0, next_timer - monotonic())):
                        break
                except TimeoutError:
                    pass
                event = events.Timer(
                    self.sender, timer=self, count=count, callback=self._callback
                )
                try:
                    await self.target.post_message(event)
                except EventTargetGone:
                    break
                count += 1
                await _wait_active()
        except CancelledError:
            pass



from __future__ import annotations

import logging

from rich.console import RenderableType
from rich.style import StyleType


from .. import events
from ..layouts.grid import GridLayout
from ..message import Message
from ..scrollbar import ScrollTo, ScrollBar
from ..geometry import clamp
from ..page import Page
from ..view import View
from ..reactive import Reactive


log = logging.getLogger("rich")


class ScrollView(View):
    def __init__(
        self,
        renderable: RenderableType | None = None,
        *,
        name: str | None = None,
        style: StyleType = "",
        fluid: bool = True,
    ) -> None:
        self.fluid = fluid
        self.vscroll = ScrollBar(vertical=True)
        self.hscroll = ScrollBar(vertical=False)
        self.page = Page(renderable or "", style=style)
        layout = GridLayout()
        layout.add_column("main")
        layout.add_column("vscroll", size=1)
        layout.add_row("main")
        layout.add_row("hscroll", size=1)
        layout.add_areas(
            content="main,main", vscroll="vscroll,main", hscroll="main,hscroll"
        )
        layout.show_row("hscroll", False)
        layout.show_row("vscroll", False)
        super().__init__(name=name, layout=layout)

    x: Reactive[float] = Reactive(0)
    y: Reactive[float] = Reactive(0)

    target_x: Reactive[float] = Reactive(0)
    target_y: Reactive[float] = Reactive(0)

    def validate_x(self, value: float) -> float:
        return clamp(value, 0, self.page.contents_size.width - self.size.width)

    def validate_target_x(self, value: float) -> float:
        return clamp(value, 0, self.page.contents_size.width - self.size.width)

    def validate_y(self, value: float) -> float:
        return clamp(value, 0, self.page.contents_size.height - self.size.height)

    def validate_target_y(self, value: float) -> float:
        return clamp(value, 0, self.page.contents_size.height - self.size.height)

    async def watch_x(self, new_value: float) -> None:
        self.page.x = round(new_value)
        self.hscroll.position = round(new_value)

    async def watch_y(self, new_value: float) -> None:
        self.page.y = round(new_value)
        self.vscroll.position = round(new_value)

    async def update(self, renderabe: RenderableType) -> None:
        self.page.update(renderabe)
        self.require_repaint()

    async def on_mount(self, event: events.Mount) -> None:
        assert isinstance(self.layout, GridLayout)
        self.layout.place(
            content=self.page,
            vscroll=self.vscroll,
            hscroll=self.hscroll,
        )
        await self.layout.mount_all(self)

    def scroll_up(self) -> None:
        self.target_y += 1.5
        self.animate("y", self.target_y, easing="out_cubic", speed=80)

    def scroll_down(self) -> None:
        self.target_y -= 1.5
        self.animate("y", self.target_y, easing="out_cubic", speed=80)

    def page_up(self) -> None:
        self.target_y -= self.size.height
        self.animate("y", self.target_y, easing="out_cubic")

    def page_down(self) -> None:
        self.target_y += self.size.height
        self.animate("y", self.target_y, easing="out_cubic")

    def page_left(self) -> None:
        self.target_x -= self.size.width
        self.animate("x", self.target_x, speed=120, easing="out_cubic")

    def page_right(self) -> None:
        self.target_x += self.size.width
        self.animate("x", self.target_x, speed=120, easing="out_cubic")

    async def on_mouse_scroll_up(self, event: events.MouseScrollUp) -> None:
        self.scroll_up()

    async def on_mouse_scroll_down(self, event: events.MouseScrollUp) -> None:
        self.scroll_down()

    async def on_key(self, event: events.Key) -> None:
        await self.dispatch_key(event)

    async def key_down(self) -> None:
        self.target_y += 2
        self.animate("y", self.target_y, easing="linear", speed=100)

    async def key_up(self) -> None:
        self.target_y -= 2
        self.animate("y", self.target_y, easing="linear", speed=100)

    async def key_pagedown(self) -> None:
        self.target_y += self.size.height
        self.animate("y", self.target_y, easing="out_cubic")

    async def key_pageup(self) -> None:
        self.target_y -= self.size.height
        self.animate("y", self.target_y, easing="out_cubic")

    async def key_end(self) -> None:
        self.target_x = 0
        self.target_y = self.page.contents_size.height - self.size.height
        self.animate("x", self.target_x, duration=1, easing="out_cubic")
        self.animate("y", self.target_y, duration=1, easing="out_cubic")

    async def key_home(self) -> None:
        self.target_x = 0
        self.target_y = 0
        self.animate("x", self.target_x, duration=1, easing="out_cubic")
        self.animate("y", self.target_y, duration=1, easing="out_cubic")

    async def on_resize(self, event: events.Resize) -> None:
        await super().on_resize(event)
        if self.fluid:
            self.page.update()

    async def message_scroll_up(self, message: Message) -> None:
        self.page_up()

    async def message_scroll_down(self, message: Message) -> None:
        self.page_down()

    async def message_scroll_left(self, message: Message) -> None:
        self.page_left()

    async def message_scroll_right(self, message: Message) -> None:
        self.page_right()

    async def message_scroll_to(self, message: ScrollTo) -> None:
        if message.x is not None:
            self.target_x = message.x
        if message.y is not None:
            self.target_y = message.y
        self.animate("x", self.target_x, speed=150, easing="out_cubic")
        self.animate("y", self.target_y, speed=150, easing="out_cubic")

    async def message_page_update(self, message: Message) -> None:
        self.x = self.validate_x(self.x)
        self.y = self.validate_y(self.y)
        self.vscroll.virtual_size = self.page.virtual_size.height
        self.vscroll.window_size = self.size.height
        if self.layout.show_column(
            "vscroll", self.page.virtual_size.height > self.size.height
        ):
            self.page.update()

        self.hscroll.virtual_size = self.page.virtual_size.width
        self.hscroll.window_size = self.size.width

        if self.layout.show_row(
            "hscroll", self.page.virtual_size.width > self.size.width
        ):
            self.page.update()



from rich.console import RenderableType
from rich.text import Text
import rich.repr

from .. import events
from ..widget import Widget


@rich.repr.auto
class Footer(Widget):
    def __init__(self) -> None:
        self.keys: list[tuple[str, str]] = []
        super().__init__()
        self.layout_size = 1

    def __rich_repr__(self) -> rich.repr.RichReprResult:
        yield "footer"

    def add_key(self, key: str, label: str) -> None:
        self.keys.append((key, label))

    def render(self) -> RenderableType:

        text = Text(
            style="white on dark_green",
            no_wrap=True,
            overflow="ellipsis",
            justify="left",
            end="",
        )
        for key, label in self.keys:
            text.append(f" {key.upper()} ", style="default on default")
            text.append(f" {label} ")
        return text



from __future__ import annotations

from abc import ABC, abstractmethod, abstractmethod
from dataclasses import dataclass
from itertools import chain
import logging
from operator import itemgetter
from time import time
from typing import cast, Iterable, Mapping, NamedTuple, TYPE_CHECKING

import rich.repr
from rich.control import Control
from rich.console import Console, ConsoleOptions, RenderResult, RenderableType
from rich.segment import Segment, Segments, SegmentLines
from rich.style import Style

from ._loop import loop_last
from ._profile import timer
from ._types import Lines

from .geometry import clamp, Region, Point

log = logging.getLogger("rich")


if TYPE_CHECKING:
    from .widget import Widget, WidgetID
    from .view import View


class NoWidget(Exception):
    pass


@rich.repr.auto
class OrderedRegion(NamedTuple):
    region: Region
    order: tuple[int, int]

    def __rich_repr__(self) -> rich.repr.RichReprResult:
        yield "region", self.region
        yield "order", self.order


class ReflowResult(NamedTuple):
    """The result of a reflow operation. Describes the chances to widgets."""

    hidden: set[Widget]
    shown: set[Widget]
    resized: set[Widget]


class LayoutUpdate:
    def __init__(self, lines: Lines, x: int, y: int) -> None:
        self.lines = lines
        self.x = x
        self.y = y

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Control.home().segment
        x = self.x
        new_line = Segment.line()
        move_to = Control.move_to
        for last, (y, line) in loop_last(enumerate(self.lines, self.y)):
            yield move_to(x, y).segment
            yield from line
            if not last:
                yield new_line


class Layout(ABC):
    """Responsible for arranging Widgets in a view."""

    def __init__(self) -> None:
        self._layout_map: dict[Widget, OrderedRegion] = {}
        self.width = 0
        self.height = 0
        self.renders: dict[Widget, tuple[Region, Lines]] = {}
        self._cuts: list[list[int]] | None = None
        self._require_update: bool = True

    def check_update(self) -> bool:
        return self._require_update

    def require_update(self) -> None:
        self._require_update = True

    def reset_update(self) -> None:
        self._require_update = False

    def reset(self) -> None:
        self._cuts = None

    def reflow(self, width: int, height: int) -> ReflowResult:
        self.reset()

        map = self.generate_map(width, height)

        # Filter out widgets that are off screen or zero area
        screen_region = Region(0, 0, width, height)
        map = {
            widget: map_region
            for widget, map_region in map.items()
            if map_region.region and screen_region.overlaps(map_region.region)
        }

        old_widgets = set(self._layout_map.keys())
        new_widgets = set(map.keys())
        # Newly visible widgets
        shown_widgets = new_widgets - old_widgets
        # Newly hidden widgets
        hidden_widgets = old_widgets - new_widgets

        self._layout_map = map
        self.width = width
        self.height = height

        # Copy renders if the size hasn't changed
        new_renders = {
            widget: (region, self.renders[widget][1])
            for widget, (region, _order) in map.items()
            if widget in self.renders
            and self.renders[widget][0].size == region.size
            and not widget.check_repaint()
        }
        self.renders = new_renders

        # Widgets with changed size
        resized_widgets = {
            widget
            for widget, (region, _order) in map.items()
            if widget in old_widgets and widget.size != region.size
        }

        return ReflowResult(
            hidden=hidden_widgets, shown=shown_widgets, resized=resized_widgets
        )

    @abstractmethod
    def get_widgets(self) -> Iterable[Widget]:
        ...

    @abstractmethod
    def generate_map(
        self, width: int, height: int, offset: Point = Point(0, 0)
    ) -> dict[Widget, OrderedRegion]:
        ...

    async def mount_all(self, view: "View") -> None:
        await view.mount(*self.get_widgets())

    @property
    def map(self) -> dict[Widget, OrderedRegion]:
        return self._layout_map

    def __iter__(self) -> Iterable[tuple[Widget, Region]]:
        layers = sorted(
            self._layout_map.items(), key=lambda item: item[1].order, reverse=True
        )
        for widget, (region, _) in layers:
            yield widget, region

    def __reversed__(self) -> Iterable[tuple[Widget, Region]]:
        layers = sorted(self._layout_map.items(), key=lambda item: item[1].order)
        for widget, (region, _) in layers:
            yield widget, region

    def get_offset(self, widget: Widget) -> Point:
        try:
            return self._layout_map[widget].region.origin
        except KeyError:
            raise NoWidget("Widget is not in layout")

    def get_widget_at(self, x: int, y: int) -> tuple[Widget, Region]:
        """Get the widget under the given point or None."""
        for widget, region in self:
            if widget.is_visual and region.contains(x, y):
                return widget, region
        raise NoWidget(f"No widget under screen coordinate ({x}, {y})")

    def get_style_at(self, x: int, y: int) -> Style:
        try:
            widget, region = self.get_widget_at(x, y)
        except NoWidget:
            return Style.null()
        _region, lines = self.renders[widget]
        x -= region.x
        y -= region.y
        line = lines[y]
        end = 0
        for segment in line:
            end += segment.cell_length
            if x < end:
                return segment.style or Style.null()
        return Style.null()

    def get_widget_region(self, widget: Widget) -> Region:
        try:
            region, _ = self._layout_map[widget]
        except KeyError:
            raise NoWidget("Widget is not in layout")
        else:
            return region

    @property
    def cuts(self) -> list[list[int]]:
        if self._cuts is not None:
            return self._cuts
        width = self.width
        height = self.height
        screen = Region(0, 0, width, height)
        cuts_sets = [{0, width} for _ in range(height)]

        for region, order in self._layout_map.values():
            region = region.clip(width, height)
            if region and region in screen:
                for y in range(region.y, region.y + region.height):
                    cuts_sets[y].update({region.x, region.x + region.width})

        # Sort the cuts for each line
        self._cuts = [sorted(cut_set) for cut_set in cuts_sets]
        return self._cuts

    def _get_renders(self, console: Console) -> Iterable[tuple[Region, Lines]]:
        width = self.width
        height = self.height
        screen_region = Region(0, 0, width, height)
        layout_map = self._layout_map

        widget_regions = sorted(
            ((widget, region, order) for widget, (region, order) in layout_map.items()),
            key=itemgetter(2),
            reverse=True,
        )

        def render(widget: Widget, width: int, height: int) -> Lines:
            lines = console.render_lines(
                widget, console.options.update_dimensions(width, height)
            )
            return lines

        for widget, region, _order in widget_regions:

            if not widget.is_visual:
                continue
            region_lines = self.renders.get(widget)
            if region_lines is not None:
                yield region_lines
                continue

            lines = render(widget, region.width, region.height)
            if region in screen_region:
                self.renders[widget] = (region, lines)
                yield region, lines
            elif screen_region.overlaps(region):
                new_region = region.clip(width, height)
                delta_x = new_region.x - region.x
                delta_y = new_region.y - region.y
                region = new_region

                splits = [delta_x, delta_x + region.width]
                divide = Segment.divide
                lines = [
                    list(divide(line, splits))[1]
                    for line in lines[delta_y : delta_y + region.height]
                ]
                self.renders[widget] = (region, lines)
                yield region, lines

    @classmethod
    def _assemble_chops(
        cls, chops: list[dict[int, list[Segment] | None]]
    ) -> Iterable[list[Segment]]:

        for bucket in chops:
            line: list[Segment] = []
            extend = line.extend
            for _, segments in sorted(bucket.items()):
                if segments:
                    extend(segments)
            yield line

    def render(
        self,
        console: Console,
        clip: Region = None,
    ) -> SegmentLines:
        """Render a layout.

        Args:
            layout_map (dict[WidgetID, MapRegion]): A layout map.
            console (Console): Console instance.
            width (int): Width
            height (int): Height

        Returns:
            SegmentLines: A renderable
        """
        width = self.width
        height = self.height
        screen = Region(0, 0, width, height)
        clip = clip or screen
        clip_x, clip_y, clip_x2, clip_y2 = clip.corners

        divide = Segment.divide
        back = Style.parse("black")

        # Maps each cut on to a list of segments
        cuts = self.cuts
        chops: list[dict[int, list[Segment] | None]] = [
            {cut: None for cut in cut_set} for cut_set in cuts
        ]

        # TODO: Provide an option to update the background
        background_render = [[Segment(" " * width, back)] for _ in range(height)]
        # Go through all the renders in reverse order and fill buckets with no render
        renders = self._get_renders(console)
        for region, lines in chain(renders, [(screen, background_render)]):
            for y, line in enumerate(lines, region.y):
                if clip_y > y > clip_y2:
                    continue
                first_cut = clamp(region.x, clip_x, clip_x2)
                last_cut = clamp(region.x + region.width, clip_x, clip_x2)
                final_cuts = [cut for cut in cuts[y] if (last_cut >= cut >= first_cut)]
                if len(final_cuts) > 1:
                    if final_cuts == [region.x, region.x + region.width]:
                        cut_segments = [line]
                    else:
                        relative_cuts = [cut - region.x for cut in final_cuts]
                        _, *cut_segments = divide(line, relative_cuts)
                    for cut, segments in zip(final_cuts, cut_segments):
                        if chops[y][cut] is None:
                            chops[y][cut] = segments

        # Assemble the cut renders in to lists of segments
        output_lines = list(self._assemble_chops(chops[clip_y:clip_y2]))
        return SegmentLines(output_lines, new_lines=True)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.render(console)

    def update_widget(self, console: Console, widget: Widget) -> LayoutUpdate | None:
        if widget not in self.renders:
            return None
        region, lines = self.renders[widget]
        new_lines = console.render_lines(
            widget, console.options.update_dimensions(region.width, region.height)
        )
        self.renders[widget] = (region, new_lines)

        update_lines = self.render(console, region).lines
        return LayoutUpdate(update_lines, region.x, region.y)



from __future__ import annotations

from abc import ABC, abstractmethod
from itertools import chain
import logging
from typing import Iterable, TYPE_CHECKING

from rich.console import Console, ConsoleOptions, RenderResult, RenderableType
import rich.repr
from rich.style import Style

from . import events
from .layout import Layout, NoWidget
from .layouts.dock import DockLayout
from .geometry import Dimensions, Point, Region
from .messages import UpdateMessage, LayoutMessage

from .widget import Widget, Widget


if TYPE_CHECKING:
    from .app import App

log = logging.getLogger("rich")


@rich.repr.auto
class View(Widget):
    def __init__(self, layout: Layout = None, name: str | None = None) -> None:
        self.layout: Layout = layout or DockLayout()
        self.mouse_over: Widget | None = None
        self.focused: Widget | None = None
        self.size = Dimensions(0, 0)
        self.widgets: set[Widget] = set()
        self.named_widgets: dict[str, Widget] = {}
        super().__init__(name)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        return
        yield

    def __rich_repr__(self) -> rich.repr.RichReprResult:
        yield "name", self.name

    def __getitem__(self, widget_name: str) -> Widget:
        return self.named_widgets[widget_name]

    @property
    def is_visual(self) -> bool:
        return False

    @property
    def is_root_view(self) -> bool:
        return self.parent is self.app

    def is_mounted(self, widget: Widget) -> bool:
        return widget in self.widgets

    def render(self) -> RenderableType:
        return self.layout

    def get_offset(self, widget: Widget) -> Point:
        return self.layout.get_offset(widget)

    async def message_update(self, message: UpdateMessage) -> None:
        widget = message.widget
        assert isinstance(widget, Widget)
        display_update = self.root_view.layout.update_widget(self.console, widget)
        if display_update is not None:
            self.app.display(display_update)

    async def message_layout(self, message: LayoutMessage) -> None:
        await self.root_view.refresh_layout()

    async def mount(self, *anon_widgets: Widget, **widgets: Widget) -> None:

        name_widgets: Iterable[tuple[str | None, Widget]]
        name_widgets = chain(
            ((None, widget) for widget in anon_widgets), widgets.items()
        )
        for name, widget in name_widgets:
            name = name or widget.name
            if name:
                self.named_widgets[name] = widget
            await self.app.register(widget)
            widget.set_parent(self)
            await widget.post_message(events.Mount(sender=self))
            self.widgets.add(widget)

        self.require_repaint()

    async def refresh_layout(self) -> None:

        if not self.size or not self.is_root_view:
            return

        width, height = self.console.size
        hidden, shown, resized = self.layout.reflow(width, height)
        self.app.refresh()

        for widget in self.layout.get_widgets():
            if not self.is_mounted(widget):
                await self.mount(widget)

        for widget in hidden:
            widget.post_message_no_wait(events.Hide(self))
        for widget in shown:
            widget.post_message_no_wait(events.Show(self))

        send_resize = shown
        send_resize.update(resized)

        for widget, region in self.layout:
            if widget in send_resize:
                widget.post_message_no_wait(
                    events.Resize(self, region.width, region.height)
                )

    async def on_resize(self, event: events.Resize) -> None:
        self.size = Dimensions(event.width, event.height)
        await self.refresh_layout()

    def get_widget_at(self, x: int, y: int) -> tuple[Widget, Region]:
        return self.layout.get_widget_at(x, y)

    def get_style_at(self, x: int, y: int) -> Style:
        return self.layout.get_style_at(x, y)

    def get_widget_region(self, widget: Widget) -> Region:
        return self.layout.get_widget_region(widget)

    async def _on_mouse_move(self, event: events.MouseMove) -> None:
        try:
            if self.app.mouse_captured:
                widget = self.app.mouse_captured
                region = self.get_widget_region(widget)
            else:
                widget, region = self.get_widget_at(event.x, event.y)
                log.debug("WIDGET %r", widget)
        except NoWidget:
            await self.app.set_mouse_over(None)
        else:
            await self.app.set_mouse_over(widget)

            await widget.forward_event(
                events.MouseMove(
                    self,
                    event.x - region.x,
                    event.y - region.y,
                    event.delta_x,
                    event.delta_y,
                    event.button,
                    event.shift,
                    event.meta,
                    event.ctrl,
                    screen_x=event.screen_x,
                    screen_y=event.screen_y,
                    style=event.style,
                )
            )

    async def forward_event(self, event: events.Event) -> None:

        if isinstance(event, (events.Enter, events.Leave)):
            await self.post_message(event)

        elif isinstance(event, events.MouseMove):
            event.style = self.get_style_at(event.screen_x, event.screen_y)
            await self._on_mouse_move(event)

        elif isinstance(event, events.MouseEvent):
            try:
                if self.app.mouse_captured:
                    widget = self.app.mouse_captured
                    region = self.get_widget_region(widget)
                else:
                    widget, region = self.get_widget_at(event.x, event.y)
            except NoWidget:
                pass
            else:
                if isinstance(event, events.MouseDown) and widget.can_focus:
                    await self.app.set_focus(widget)
                event.style = self.get_style_at(event.screen_x, event.screen_y)
                await widget.forward_event(event.offset(-region.x, -region.y))

        elif isinstance(event, (events.MouseScrollDown, events.MouseScrollUp)):
            try:
                widget, _region = self.get_widget_at(event.x, event.y)
            except NoWidget:
                return
            scroll_widget = widget or self.focused
            if scroll_widget is not None:
                await scroll_widget.forward_event(event)
        else:
            if self.focused is not None:
                await self.focused.forward_event(event)

    async def action_toggle(self, name: str) -> None:
        widget = self.named_widgets[name]
        widget.visible = not widget.visible
        await self.post_message(LayoutMessage(self))
        # await self.refresh_layout()



from __future__ import annotations

from typing import Any, NamedTuple, TypeVar


T = TypeVar("T", int, float)


def clamp(value: T, minimum: T, maximum: T) -> T:
    """Clamps a value between two other values.

    Args:
        value (T): A value
        minimum (T): Minimum value
        maximum (T): maximum value

    Returns:
        T: New value that is not less than the minimum or greater than the maximum.
    """
    if value < minimum:
        return minimum
    elif value > maximum:
        return maximum
    else:
        return value


class Point(NamedTuple):
    """A point defined by x and y coordinates."""

    x: int
    y: int

    @property
    def is_origin(self) -> bool:
        """Check if the point is at the origin (0, 0)"""
        return self == (0, 0)

    def __add__(self, other: object) -> Point:
        if isinstance(other, tuple):
            _x, _y = self
            x, y = other
            return Point(_x + x, _y + y)
        return NotImplemented

    def __sub__(self, other: object) -> Point:
        if isinstance(other, tuple):
            _x, _y = self
            x, y = other
            return Point(_x - x, _y - y)
        return NotImplemented

    def blend(self, destination: Point, factor: float) -> Point:
        """Blend (interpolate) to a new point.

        Args:
            destination (Point): Point where progress is 1.0
            factor (float): A value between 0 and 1.0

        Returns:
            Point: A new point on a line between self and destination
        """
        x1, y1 = self
        x2, y2 = destination
        return Point(int(x1 + (x2 - x1) * factor), int((y1 + (y2 - y1) * factor)))


class Dimensions(NamedTuple):
    """An area defined by its width and height."""

    width: int
    height: int

    def __bool__(self) -> bool:
        return self.width * self.height != 0

    @property
    def area(self) -> int:
        return self.width * self.height

    def contains(self, x: int, y: int) -> bool:
        """Check if a point is in the region.

        Args:
            x (int): X coordinate (column)
            y (int): Y coordinate (row)

        Returns:
            bool: True if the point is within the region.
        """
        width, height = self
        return width > x >= 0 and height > y >= 0

    def contains_point(self, point: tuple[int, int]) -> bool:
        """Check if a point is in the region.

        Args:
            point (tuple[int, int]): A tuple of x and y coordinates.

        Returns:
            bool: True if the point is within the region.
        """
        x, y = point
        width, height = self
        return width > x >= 0 and height > y >= 0

    def __contains__(self, other: Any) -> bool:
        try:
            x, y = other
        except Exception:
            raise TypeError(
                "Dimensions.__contains__ requires an iterable of two integers"
            )
        width, height = self
        return width > x >= 0 and height > y >= 0


class Region(NamedTuple):
    """Defines a rectangular region of the screen."""

    x: int
    y: int
    width: int
    height: int

    @classmethod
    def from_corners(cls, x1: int, y1: int, x2: int, y2: int) -> Region:
        """Construct a Region form the top left and bottom right corners.

        Args:
            x1 (int): Top left x
            y1 (int): Top left y
            x2 (int): Bottom right x
            y2 (int): Bottom right y

        Returns:
            Region: A new region.
        """
        return cls(x1, y1, x2 - x1, y2 - y1)

    def __bool__(self) -> bool:
        return self.width != 0 and self.height != 0

    @property
    def area(self) -> int:
        """Get the area within the region."""
        return self.width * self.height

    @property
    def origin(self) -> Point:
        """Get the start point of the region."""
        return Point(self.x, self.y)

    @property
    def limit(self) -> Point:
        x, y, width, height = self
        return Point(x + width, y + height)

    @property
    def limit_inclusive(self) -> Point:
        """Get the end point of the region."""
        x, y, width, height = self
        return Point(x + width - 1, y + height - 1)

    @property
    def size(self) -> Dimensions:
        """Get the size of the region."""
        return Dimensions(self.width, self.height)

    @property
    def corners(self) -> tuple[int, int, int, int]:
        """Get the maxima and minima of region.

        Returns:
            tuple[int, int, int, int]: A tuple of (<min x>, <max x>, <min y>, <max y>)
        """
        x, y, width, height = self
        return x, y, x + width, y + height

    def __add__(self, other: Any) -> Region:
        if isinstance(other, tuple):
            ox, oy = other
            x, y, width, height = self
            return Region(x + ox, y + oy, width, height)
        return NotImplemented

    def overlaps(self, other: Region) -> bool:
        """Check if another region overlaps this region.

        Args:
            other (Region): A Region.

        Returns:
            bool: True if other region shares any cells with this region.
        """
        x, y, x2, y2 = self.corners
        ox, oy, ox2, oy2 = other.corners

        return ((x2 > ox >= x) or (x2 > ox2 > x) or (ox < x and ox2 > x2)) and (
            (y2 > oy >= y) or (y2 > oy2 > y) or (oy < y and oy2 > y2)
        )

    def contains(self, x: int, y: int) -> bool:
        """Check if a point is in the region.

        Args:
            x (int): X coordinate (column)
            y (int): Y coordinate (row)

        Returns:
            bool: True if the point is within the region.
        """
        self_x, self_y, width, height = self
        return (self_x + width > x >= self_x) and (self_y + height > y >= self_y)

    def contains_point(self, point: tuple[int, int]) -> bool:
        """Check if a point is in the region.

        Args:
            point (tuple[int, int]): A tuple of x and y coordinates.

        Returns:
            bool: True if the point is within the region.
        """
        x1, y1, x2, y2 = self.corners
        try:
            ox, oy = point
        except Exception:
            raise TypeError(f"a tuple of two integers is required, not {point!r}")
        return (x2 > ox >= x1) and (y2 > oy >= y1)

    def contains_region(self, other: Region) -> bool:
        """Check if a region is entirely contained within this region.

        Args:
            other (Region): A region.

        Returns:
            bool: True if the other region fits perfectly within this region.
        """
        x1, y1, x2, y2 = self.corners
        ox, oy, ox2, oy2 = other.corners
        return (x2 >= ox >= x1 and y2 >= oy >= y1) and (
            x2 >= ox2 >= x1 and y2 >= oy2 >= y1
        )

    def translate(self, translate_x: int, translate_y: int) -> Region:
        """Move the origin of the Region.

        Args:
            translate_x (int): Value to add to x coordinate.
            translate_y (int): Value to add to y coordinate.

        Returns:
            Region: A new region shifted by x, y
        """

        x, y, width, height = self
        return Region(x + translate_x, y + translate_y, width, height)

    def __contains__(self, other: Any) -> bool:
        """Check if a point is in this region."""
        if isinstance(other, Region):
            return self.contains_region(other)
        else:
            try:
                return self.contains_point(other)
            except TypeError:
                return False

    def clip(self, width: int, height: int) -> Region:
        """Clip this region to fit within width, height.

        Args:
            width (int): Width of bounds.
            height (int): Height of bounds.

        Returns:
            Region: Clipped region.
        """
        x1, y1, x2, y2 = self.corners

        new_region = Region.from_corners(
            clamp(x1, 0, width),
            clamp(y1, 0, height),
            clamp(x2, 0, width),
            clamp(y2, 0, height),
        )
        return new_region

    def clip_region(self, region: Region) -> Region:
        """Clip this region to fit within another region.

        Args:
            region ([type]): A region that overlaps this region.

        Returns:
            Region: A new region that fits within ``region``.
        """
        x1, y1, x2, y2 = self.corners
        cx1, cy1, cx2, cy2 = region.corners

        new_region = Region.from_corners(
            clamp(x1, cx1, cx2),
            clamp(y1, cy1, cy2),
            clamp(x2, cx2, cx2),
            clamp(y2, cy2, cy2),
        )
        return new_region


if __name__ == "__main__":
    from rich import print

    region = Region(-5, -5, 60, 100)

    print(region.clip(80, 25))

    region = Region(10, 10, 90, 90)

    print(region.corners)

    print((15, 15) in region)
    print((5, 15) in region)
    print(Region(15, 15, 10, 10) in region)



from __future__ import annotations

from functools import partial
import sys
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Type,
    Union,
    TypeVar,
    TYPE_CHECKING,
)
from weakref import WeakSet

from . import events

from .message_pump import MessagePump
from ._types import MessageTarget

if TYPE_CHECKING:
    from .message import Message
    from .app import App
    from .widget import Widget

    Reactable = Union[Widget, App]

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


ReactiveType = TypeVar("ReactiveType")


class Reactive(Generic[ReactiveType]):
    """Reactive descriptor."""

    def __init__(
        self,
        default: ReactiveType,
        *,
        layout: bool = False,
        repaint: bool = True,
    ) -> None:
        self._default = default
        self.layout = layout
        self.repaint = repaint
        self._first = True

    def __set_name__(self, owner: Type[MessageTarget], name: str) -> None:
        self.name = name
        self.internal_name = f"__{name}"
        setattr(owner, self.internal_name, self._default)

    def __get__(self, obj: Reactable, obj_type: type[object]) -> ReactiveType:
        return getattr(obj, self.internal_name)

    def __set__(self, obj: Reactable, value: ReactiveType) -> None:

        name = self.name
        internal_name = f"__{name}"
        current_value = getattr(obj, internal_name, None)
        validate_function = getattr(obj, f"validate_{name}", None)
        if callable(validate_function):
            value = validate_function(value)

        if current_value != value or self._first:
            self._first = False
            setattr(obj, internal_name, value)

            self.check_watchers(obj, name)

            if self.layout:
                obj.require_layout()
            elif self.repaint:
                obj.require_repaint()

    @classmethod
    def check_watchers(cls, obj: Reactable, name: str) -> None:

        internal_name = f"__{name}"
        value = getattr(obj, internal_name)

        watch_function = getattr(obj, f"watch_{name}", None)
        if callable(watch_function):
            obj.post_message_no_wait(
                events.Callback(obj, callback=partial(watch_function, value))
            )

        watcher_name = f"__{name}_watchers"
        watchers = getattr(obj, watcher_name, ())
        for watcher in watchers:
            obj.post_message_no_wait(
                events.Callback(obj, callback=partial(watcher, value))
            )


def watch(
    obj: Reactable, attribute_name: str, callback: Callable[[Any], Awaitable]
) -> None:
    watcher_name = f"__{attribute_name}_watchers"
    if not hasattr(obj, watcher_name):
        setattr(obj, watcher_name, WeakSet())
    watchers = getattr(obj, watcher_name)
    watchers.add(callback)
    Reactive.check_watchers(obj, attribute_name)



from __future__ import annotations

from functools import partial
from logging import getLogger
from typing import (
    Any,
    TYPE_CHECKING,
    Callable,
    ClassVar,
    NewType,
    cast,
)
from weakref import WeakValueDictionary

import rich.repr
from rich.align import Align
from rich.console import Console, RenderableType
from rich.panel import Panel
from rich.pretty import Pretty
from rich.style import Style

from . import events
from ._animator import BoundAnimator
from ._context import active_app
from .geometry import Dimensions
from .message import Message
from .message_pump import MessagePump
from .messages import LayoutMessage, UpdateMessage
from .reactive import Reactive, watch

if TYPE_CHECKING:
    from .app import App
    from .view import View


WidgetID = NewType("WidgetID", int)

log = getLogger("rich")


@rich.repr.auto
class Widget(MessagePump):
    _id: ClassVar[int] = 0
    _counts: ClassVar[dict[str, int]] = {}
    can_focus: bool = False

    def __init__(self, name: str | None = None) -> None:
        class_name = self.__class__.__name__
        Widget._counts.setdefault(class_name, 0)
        Widget._counts[class_name] += 1
        _count = self._counts[class_name]
        self.id: WidgetID = cast(WidgetID, Widget._id)
        Widget._id += 1

        self.name = name or f"{class_name}#{_count}"

        self.size = Dimensions(0, 0)
        self.size_changed = False
        self._repaint_required = False
        self._layout_required = False
        self._animate: BoundAnimator | None = None
        self._reactive_watches: dict[str, Callable] = {}

        super().__init__()

    visible: Reactive[bool] = Reactive(True, layout=True)
    layout_size: Reactive[int | None] = Reactive(None)
    layout_fraction: Reactive[int] = Reactive(1)
    layout_min_size: Reactive[int] = Reactive(1)
    layout_offset_x: Reactive[float] = Reactive(0, layout=True)
    layout_offset_y: Reactive[float] = Reactive(0, layout=True)

    def __init_subclass__(cls, can_focus: bool = True) -> None:
        super().__init_subclass__()
        cls.can_focus = can_focus

    def __rich_repr__(self) -> rich.repr.RichReprResult:
        yield "name", self.name

    def __rich__(self) -> RenderableType:
        return self.render()

    def watch(self, attribute_name, callback: Callable[[Any], None]) -> None:
        watch(self, attribute_name, callback)

    @property
    def is_visual(self) -> bool:
        return True

    @property
    def console(self) -> Console:
        """Get the current console."""
        return active_app.get().console

    @property
    def root_view(self) -> "View":
        """Return the top-most view."""
        return active_app.get().view

    @property
    def animate(self) -> BoundAnimator:
        if self._animate is None:
            self._animate = self.app.animator.bind(self)
        assert self._animate is not None
        return self._animate

    @property
    def layout_offset(self) -> tuple[int, int]:
        """Get the layout offset as a tuple."""
        return (round(self.layout_offset_x), round(self.layout_offset_y))

    def require_repaint(self) -> None:
        """Mark widget as requiring a repaint.

        Actual repaint is done by parent on idle.
        """
        self._repaint_required = True
        self.post_message_no_wait(events.Null(self))

    def require_layout(self) -> None:
        self._layout_required = True
        self.post_message_no_wait(events.Null(self))

    def check_repaint(self) -> bool:
        return self._repaint_required

    def check_layout(self) -> bool:
        return self._layout_required

    def reset_check_repaint(self) -> None:
        self._repaint_required = False

    def reset_check_layout(self) -> None:
        self._layout_required = False

    def get_style_at(self, x: int, y: int) -> Style:
        offset_x, offset_y = self.root_view.get_offset(self)
        return self.root_view.get_style_at(x + offset_x, y + offset_y)

    async def call_later(self, callback: Callable, *args, **kwargs) -> None:
        await self.app.call_later(callback, *args, **kwargs)

    async def forward_event(self, event: events.Event) -> None:
        await self.post_message(event)

    async def refresh(self) -> None:
        """Re-render the window and repaint it."""
        self.require_repaint()
        await self.repaint()

    async def repaint(self) -> None:
        """Instructs parent to repaint this widget."""
        await self.emit(UpdateMessage(self, self))

    async def update_layout(self) -> None:
        await self.emit(LayoutMessage(self))

    def render(self) -> RenderableType:
        """Get renderable for widget.

        Returns:
            RenderableType: Any renderable
        """
        return Panel(
            Align.center(Pretty(self), vertical="middle"), title=self.__class__.__name__
        )

    async def action(self, action: str, *params) -> None:
        await self.app.action(action, self)

    async def post_message(self, message: Message) -> bool:
        if not self.check_message_enabled(message):
            return True
        return await super().post_message(message)

    async def on_event(self, event: events.Event) -> None:
        if isinstance(event, events.Resize):
            new_size = Dimensions(event.width, event.height)
            if self.size != new_size:
                self.size = new_size
                self.require_repaint()
        await super().on_event(event)

    async def on_idle(self, event: events.Idle) -> None:
        if self.check_layout():
            self.reset_check_repaint()
            self.reset_check_layout()
            await self.update_layout()
        elif self.check_repaint():
            self.reset_check_repaint()
            self.reset_check_layout()
            await self.repaint()

    async def focus(self) -> None:
        await self.app.set_focus(self)

    async def capture_mouse(self, capture: bool = True) -> None:
        await self.app.capture_mouse(self if capture else None)

    async def release_mouse(self) -> None:
        await self.app.capture_mouse(None)

    async def broker_event(self, event_name: str, event: events.Event) -> bool:
        return await self.app.broker_event(event_name, event, default_namespace=self)

    async def dispatch_key(self, event: events.Key) -> None:
        """Dispatch a key event to method.

        This method will call the method named 'key_<event.key>' if it exists.

        Args:
            event (events.Key): A key event.
        """

        key_method = getattr(self, f"key_{event.key}", None)
        if key_method is not None:
            await key_method()

    async def on_mouse_down(self, event: events.MouseUp) -> None:
        await self.broker_event("mouse.down", event)

    async def on_mouse_up(self, event: events.MouseUp) -> None:
        await self.broker_event("mouse.up", event)

    async def on_click(self, event: events.Click) -> None:
        await self.broker_event("click", event)



from typing import Generic, Type, TypeVar


ParentType = TypeVar("ParentType")
ValueType = TypeVar("ValueType")


class Reactive(Generic[ValueType]):
    def __init__(self, default: ValueType) -> None:
        self._default = default

    def __set_name__(self, owner: object, name: str) -> None:
        self.internal_name = f"__{name}"
        setattr(owner, self.internal_name, self._default)

    def __get__(self, obj: object, obj_type: Type[object]) -> ValueType:
        print("__get__", obj, obj_type)

        return getattr(obj, self.internal_name)

    def __set__(self, obj: object, value: ValueType) -> None:
        print("__set__", obj, value)
        setattr(obj, self.internal_name, value)


class Example:
    def __init__(self, foo: int = 3) -> None:
        self.foo = foo

    color: Reactive[str] = Reactive("blue")


example = Example()

print(example.color)
example.color = "red"
print(example.color)
print(example.foo)



from time import monotonic
from typing import ClassVar

import rich.repr

from .case import camel_to_snake
from ._types import MessageTarget


@rich.repr.auto
class Message:
    """Base class for a message."""

    __slots__ = [
        "sender",
        "name",
        "time",
        "_no_default_action",
        "_stop_propagation",
    ]

    sender: MessageTarget
    bubble: ClassVar[bool] = False

    def __init__(self, sender: MessageTarget) -> None:
        self.sender = sender
        self.name = camel_to_snake(self.__class__.__name__.replace("Message", ""))
        self.time = monotonic()
        self._no_default_action = False
        self._stop_propagation = False
        super().__init__()

    def __rich_repr__(self) -> rich.repr.RichReprResult:
        yield self.sender

    def __init_subclass__(cls, bubble: bool = False) -> None:
        super().__init_subclass__()
        cls.bubble = bubble

    def can_replace(self, message: "Message") -> bool:
        """Check if another message may supersede this one.

        Args:
            message (Message): [description]

        Returns:
            bool: [description]
        """
        return False

    def prevent_default(self, prevent: bool = True) -> None:
        """Suppress the default action.

        Args:
            prevent (bool, optional): True if the default action should be suppressed,
                or False if the default actions should be performed. Defaults to True.
        """
        self._no_default_action = prevent

    def stop(self, stop: bool = True) -> None:
        self._stop_propagation = stop



from __future__ import annotations
from typing import cast, Optional

from ..layouts.dock import DockLayout, Dock, DockEdge
from ..view import View
from ..widget import Widget


class DoNotSet:
    pass


do_not_set = DoNotSet()


class DockView(View):
    def __init__(self, name: str | None = None) -> None:
        super().__init__(layout=DockLayout(), name=name)

    async def dock(
        self,
        *widgets: Widget,
        edge: DockEdge = "top",
        z: int = 0,
        size: int | None | DoNotSet = do_not_set,
        name: str | None = None
    ) -> None:

        dock = Dock(edge, widgets, z)
        assert isinstance(self.layout, DockLayout)
        self.layout.docks.append(dock)
        for widget in widgets:
            if size is not do_not_set:
                widget.layout_size = cast(Optional[int], size)
            if not self.is_mounted(widget):
                if name is None:
                    await self.mount(widget)
                else:
                    await self.mount(**{name: widget})
        await self.refresh_layout()



from __future__ import annotations
import os

import asyncio
from functools import partial
import logging
import signal
from typing import Any, Callable, ClassVar, Type, TypeVar
import warnings

from rich.control import Control
import rich.repr
from rich.screen import Screen
from rich import get_console
from rich.console import Console, RenderableType
from rich.style import Style
from rich.traceback import Traceback

from . import events
from . import actions
from ._animator import Animator
from .binding import Bindings, NoBinding
from .geometry import Point, Region
from ._context import active_app
from ._event_broker import extract_handler_actions, NoHandler
from ._types import MessageTarget
from .driver import Driver
from .layouts.dock import DockLayout, Dock
from ._linux_driver import LinuxDriver
from .message_pump import MessagePump
from .message import Message
from .view import View
from .views import DockView
from .widget import Widget, Widget, Reactive

log = logging.getLogger("rich")


# asyncio will warn against resources not being cleared
warnings.simplefilter("always", ResourceWarning)
# https://github.com/boto/boto3/issues/454
# warnings.filterwarnings(
#     "ignore", category=ResourceWarning, message="unclosed.*<ssl.SSLSocket.*>"
# )


LayoutDefinition = "dict[str, Any]"

ViewType = TypeVar("ViewType", bound=View)

# try:
#     import uvloop
# except ImportError:
#     pass
# else:
#     uvloop.install()


class PanicMessage(Message):
    def __init__(self, sender: MessageTarget, traceback: Traceback) -> None:
        self.traceback = traceback
        super().__init__(sender)


class ActionError(Exception):
    pass


class ShutdownError(Exception):
    pass


@rich.repr.auto
class App(MessagePump):

    KEYS: ClassVar[dict[str, str]] = {}

    def __init__(
        self,
        console: Console | None = None,
        screen: bool = True,
        driver_class: Type[Driver] | None = None,
        title: str = "Textual Application",
    ):
        """The Textual Application base class

        Args:
            console (Console, optional): A Rich Console. Defaults to None.
            screen (bool, optional): Enable full-screen application mode. Defaults to True.
            driver_class (Type[Driver], optional): Driver class, or None to use default. Defaults to None.
            title (str, optional): Title of the application. Defaults to "Textual Application".
        """
        self.console = console or get_console()
        self._screen = screen
        self.driver_class = driver_class or LinuxDriver
        self._title = title
        self._layout = DockLayout()
        self._view_stack: list[View] = []
        self.children: set[MessagePump] = set()

        self.focused: Widget | None = None
        self.mouse_over: Widget | None = None
        self.mouse_captured: Widget | None = None
        self._driver: Driver | None = None
        self._tracebacks: list[Traceback] = []

        self._docks: list[Dock] = []
        self._action_targets = {"app", "view"}
        self._animator = Animator(self)
        self.animate = self._animator.bind(self)
        self.mouse_position = Point(0, 0)
        self.bindings = Bindings()
        self._title = title

        self.bindings.bind("ctrl+c", "quit")

        super().__init__()

    title: Reactive[str] = Reactive("Textual")
    sub_title: Reactive[str] = Reactive("")

    def __rich_repr__(self) -> rich.repr.RichReprResult:
        yield "title", self.title

    def __rich__(self) -> RenderableType:
        return self.view

    @property
    def animator(self) -> Animator:
        return self._animator

    @property
    def view(self) -> View:
        return self._view_stack[-1]

    async def bind(
        self, keys: str, action: str, description: str = "", show: bool = False
    ) -> None:
        self.bindings.bind(keys, action, description, show=show)

    @classmethod
    def run(
        cls,
        console: Console = None,
        screen: bool = True,
        driver: Type[Driver] = None,
        **kwargs,
    ):
        """Run the app.

        Args:
            console (Console, optional): Console object. Defaults to None.
            screen (bool, optional): Enable application mode. Defaults to True.
            driver (Type[Driver], optional): Driver class or None for default. Defaults to None.
        """

        async def run_app() -> None:
            app = cls(console=console, screen=screen, driver_class=driver, **kwargs)
            await app.process_messages()

        asyncio.run(run_app())

    async def push_view(self, view: ViewType) -> ViewType:
        await self.register(view)
        view.set_parent(self)
        self._view_stack.append(view)
        return view

    def on_keyboard_interupt(self) -> None:
        loop = asyncio.get_event_loop()
        event = events.ShutdownRequest(sender=self)
        asyncio.run_coroutine_threadsafe(self.post_message(event), loop=loop)

    async def set_focus(self, widget: Widget | None) -> None:
        log.debug("set_focus %r", widget)
        if widget == self.focused:
            return

        if widget is None:
            if self.focused is not None:
                focused = self.focused
                self.focused = None
                await focused.post_message(events.Blur(self))
        elif widget.can_focus:
            if self.focused is not None:
                await self.focused.post_message(events.Blur(self))
            if widget is not None and self.focused != widget:
                self.focused = widget
                await widget.post_message(events.Focus(self))

    async def set_mouse_over(self, widget: Widget | None) -> None:
        if widget is None:
            if self.mouse_over is not None:
                try:
                    await self.mouse_over.post_message(events.Leave(self))
                finally:
                    self.mouse_over = None
        else:
            if self.mouse_over != widget:
                try:
                    if self.mouse_over is not None:
                        await self.mouse_over.forward_event(events.Leave(self))
                    if widget is not None:
                        await widget.forward_event(events.Enter(self))
                finally:
                    self.mouse_over = widget

    async def capture_mouse(self, widget: Widget | None) -> None:
        """Send all Mouse events to a given widget."""
        if widget == self.mouse_captured:
            return
        if self.mouse_captured is not None:
            await self.mouse_captured.post_message(
                events.MouseReleased(self, self.mouse_position)
            )
        self.mouse_captured = widget
        if widget is not None:
            await widget.post_message(events.MouseCaptured(self, self.mouse_position))

    def panic(self, traceback: Traceback) -> None:
        self._tracebacks.append(traceback)
        self.close_messages_no_wait()

    async def process_messages(self) -> None:
        log.debug("driver=%r", self.driver_class)
        driver = self._driver = self.driver_class(self.console, self)
        active_app.set(self)

        await self.push_view(View())

        self.view.set_parent(self)
        await self.register(self.view)

        await self.dispatch_message(events.Load(sender=self))

        try:
            driver.start_application_mode()
        except Exception:
            self.console.print_exception()
            log.exception("error starting application mode")
        else:
            traceback: Traceback | None = None

            try:
                self.title = self._title
                await self.post_message(events.Startup(sender=self))
                self.require_layout()
                await self.animator.start()

                await super().process_messages()
                await self.animator.stop()

                while self.children:
                    child = self.children.pop()
                    log.debug("closing %r", child)
                    await child.close_messages()

                while self._view_stack:
                    view = self._view_stack.pop()
                    await view.close_messages()
            except Exception:
                self._tracebacks.append(Traceback(show_locals=True))
            finally:
                driver.stop_application_mode()
                for traceback in self._tracebacks:
                    self.console.print(traceback)

    def require_repaint(self) -> None:
        self.refresh()

    def require_layout(self) -> None:
        self.view.require_layout()

    async def call_later(self, callback: Callable, *args, **kwargs) -> None:
        await self.post_message(events.Idle(self))
        await self.post_message(
            events.Callback(self, partial(callback, *args, **kwargs))
        )

    async def message_update(self, message: Message) -> None:
        self.refresh()

    async def register(self, child: MessagePump) -> None:
        self.children.add(child)
        child.start_messages()
        await child.post_message(events.Created(sender=self))

    async def remove(self, child: MessagePump) -> None:
        self.children.remove(child)

    async def shutdown(self):
        driver = self._driver
        assert driver is not None
        driver.disable_input()
        await self.close_messages()

    def refresh(self) -> None:
        sync_available = os.environ.get("TERM_PROGRAM", "") != "Apple_Terminal"
        if not self._closed:
            console = self.console
            try:
                if sync_available:
                    console.file.write("\x1bP=1s\x1b\\")
                with console:
                    console.print(Screen(Control.home(), self.view, Control.home()))
                if sync_available:
                    console.file.write("\x1bP=2s\x1b\\")
            except Exception:
                self.panic(Traceback(show_locals=True))

    def display(self, renderable: RenderableType) -> None:
        if not self._closed:
            console = self.console
            try:
                with console:
                    console.print(renderable)
            except Exception:
                log.exception("display failed")

    def get_widget_at(self, x: int, y: int) -> tuple[Widget, Region]:
        return self.view.get_widget_at(x, y)

    async def on_event(self, event: events.Event) -> None:
        if isinstance(event, events.Key):
            try:
                binding = self.bindings.get_key(event.key)
            except NoBinding:
                pass
            else:
                await self.action(binding.action)
                return

        if isinstance(event, events.InputEvent):
            if isinstance(event, events.MouseEvent):
                self.mouse_position = Point(event.x, event.y)
            if isinstance(event, events.Key) and self.focused is not None:
                await self.focused.forward_event(event)
            await self.view.forward_event(event)
        else:
            await super().on_event(event)

    async def action(
        self,
        action: str,
        default_namespace: object | None = None,
        modifiers: set[str] | None = None,
    ) -> None:
        """Perform an action.

        Args:
            action (str): Action encoded in a string.
        """
        target, params = actions.parse(action)
        if "." in target:
            destination, action_name = target.split(".", 1)
            if destination not in self._action_targets:
                raise ActionError("Action namespace {destination} is not known")
            action_target = getattr(self, destination)
        else:
            action_target = default_namespace or self
            action_name = action

        log.debug("ACTION %r %r", action_target, action_name)
        await self.dispatch_action(action_target, action_name, params)

    async def dispatch_action(
        self, namespace: object, action_name: str, params: Any
    ) -> None:
        _rich_traceback_guard = True
        method_name = f"action_{action_name}"
        method = getattr(namespace, method_name, None)
        if method is not None:
            await method(*params)

    async def broker_event(
        self, event_name: str, event: events.Event, default_namespace: object | None
    ) -> bool:
        try:
            style = getattr(event, "style")
        except AttributeError:
            return False
        try:
            modifiers, action = extract_handler_actions(event_name, style.meta)
        except NoHandler:
            return False
        await self.action(
            action, default_namespace=default_namespace, modifiers=modifiers
        )
        return True

    async def on_shutdown_request(self, event: events.ShutdownRequest) -> None:
        log.debug("shutdown request")
        await self.close_messages()

    async def on_resize(self, event: events.Resize) -> None:
        await self.view.post_message(event)

    async def action_quit(self) -> None:
        await self.shutdown()

    async def action_bang(self) -> None:
        1 / 0

    async def action_bell(self) -> None:
        self.console.bell()


if __name__ == "__main__":
    import asyncio
    from logging import FileHandler

    from rich.panel import Panel

    from .widgets import Header
    from .widgets import Footer

    from .widgets import Placeholder
    from .scrollbar import ScrollBar

    from rich.markdown import Markdown

    # from .widgets.scroll_view import ScrollView

    import os

    # from rich.console import Console

    # console = Console()
    # console.print(scroll_bar, height=10)
    # console.print(scroll_view, height=20)

    # import sys

    # sys.exit()

    logging.basicConfig(
        level="NOTSET",
        format="%(message)s",
        datefmt="[%X]",
        handlers=[FileHandler("richtui.log")],
    )

    class MyApp(App):
        """Just a test app."""

        async def on_load(self, event: events.Load) -> None:
            await self.bind("q,ctrl+c", "quit")
            await self.bind("x", "bang")
            await self.bind("b", "toggle_sidebar")

        show_bar: Reactive[bool] = Reactive(False)

        async def watch_show_bar(self, show_bar: bool) -> None:
            self.animator.animate(self.bar, "layout_offset_x", -40 if show_bar else 0)

        async def action_toggle_sidebar(self) -> None:
            self.show_bar = not self.show_bar

        async def on_startup(self, event: events.Startup) -> None:

            view = await self.push_view(DockView())

            header = Header()
            footer = Footer()
            self.bar = Placeholder(name="left")
            footer.add_key("b", "Toggle sidebar")
            footer.add_key("q", "Quit")

            await view.dock(header, edge="top")
            await view.dock(footer, edge="bottom")
            await view.dock(self.bar, edge="left", size=40, z=1)

            # await view.dock(Placeholder(), Placeholder(), edge="top")

            sub_view = DockView()
            await sub_view.dock(Placeholder(), Placeholder(), edge="top")
            await view.dock(sub_view, edge="left")

            # self.refresh()

            # footer = Footer()
            # footer.add_key("b", "Toggle sidebar")
            # footer.add_key("q", "Quit")

            # readme_path = os.path.join(
            #     os.path.dirname(os.path.abspath(__file__)), "richreadme.md"
            # )
            # # scroll_view = LayoutView()
            # # scroll_bar = ScrollBar()
            # with open(readme_path, "rt") as fh:
            #     readme = Markdown(fh.read(), hyperlinks=True, code_theme="fruity")
            # # scroll_view.layout.split_column(
            # #     Layout(readme, ratio=1), Layout(scroll_bar, ratio=2, size=2)
            # # )
            # layout = Layout()
            # layout.split_column(Layout(name="l1"), Layout(name="l2"))
            # # sub_view = LayoutView(name="Sub view", layout=layout)

            # sub_view = ScrollView(readme)

            # # await sub_view.mount_all(l1=Placeholder(), l2=Placeholder())

            # await self.view.mount_all(
            #     header=Header(self.title),
            #     left=Placeholder(),
            #     body=sub_view,
            #     footer=footer,
            # )

    # app = MyApp()
    # from rich.console import Console

    # console = Console()
    # console.print(app._view_stack[0], height=30)
    # console.print(app._view_stack)
    MyApp.run()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['textual', 'textual.layouts', 'textual.views', 'textual.widgets']

package_data = \
{'': ['*']}

install_requires = \
['rich>=10.6.0,<11.0.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.10.0,<4.0.0']}

setup_kwargs = {
    'name': 'textual',
    'version': '0.1.7',
    'description': 'Text User Interface using Rich',
    'long_description': None,
    'author': 'Will McGugan',
    'author_email': 'willmcgugan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/willmcgugan/textual',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import asyncio
import os
from codecs import getincrementaldecoder
import selectors
import signal
import sys
import logging
import termios
from time import time
import tty
from typing import Any, TYPE_CHECKING
from threading import Event, Thread

if TYPE_CHECKING:
    from rich.console import Console


from . import events
from .driver import Driver
from ._types import MessageTarget
from ._xterm_parser import XTermParser


log = logging.getLogger("rich")


class LinuxDriver(Driver):
    def __init__(self, console: "Console", target: "MessageTarget") -> None:
        super().__init__(console, target)
        self.fileno = sys.stdin.fileno()
        self.attrs_before: list[Any] | None = None
        self.exit_event = Event()
        self._key_thread: Thread | None = None

    def _get_terminal_size(self) -> tuple[int, int]:
        width: int | None = 80
        height: int | None = 25
        try:
            width, height = os.get_terminal_size(sys.stdin.fileno())
        except (AttributeError, ValueError, OSError):
            try:
                width, height = os.get_terminal_size(sys.stdout.fileno())
            except (AttributeError, ValueError, OSError):
                pass
        width = width or 80
        height = height or 25
        return width, height

    def _enable_mouse_support(self) -> None:
        write = self.console.file.write
        write("\x1b[?1000h")

        write("\x1b[?1015h")
        write("\x1b[?1006h")

        # write("\x1b[?1007h")
        self.console.file.flush()

        # Note: E.g. lxterminal understands 1000h, but not the urxvt or sgr
        #       extensions.

    def _disable_mouse_support(self) -> None:
        write = self.console.file.write
        write("\x1b[?1000l")
        write("\x1b[?1015l")
        write("\x1b[?1006l")
        self.console.file.flush()

    def start_application_mode(self):

        loop = asyncio.get_event_loop()

        def on_terminal_resize(signum, stack) -> None:
            terminal_size = self._get_terminal_size()
            width, height = terminal_size
            event = events.Resize(self._target, width, height)
            self.console.size = terminal_size
            asyncio.run_coroutine_threadsafe(
                self._target.post_message(event),
                loop=loop,
            )

        signal.signal(signal.SIGWINCH, on_terminal_resize)

        self.console.set_alt_screen(True)
        self._enable_mouse_support()
        try:
            self.attrs_before = termios.tcgetattr(self.fileno)
        except termios.error:
            # Ignore attribute errors.
            self.attrs_before = None

        try:
            newattr = termios.tcgetattr(self.fileno)
        except termios.error:
            pass
        else:

            newattr[tty.LFLAG] = self._patch_lflag(newattr[tty.LFLAG])
            newattr[tty.IFLAG] = self._patch_iflag(newattr[tty.IFLAG])

            # VMIN defines the number of characters read at a time in
            # non-canonical mode. It seems to default to 1 on Linux, but on
            # Solaris and derived operating systems it defaults to 4. (This is
            # because the VMIN slot is the same as the VEOF slot, which
            # defaults to ASCII EOT = Ctrl-D = 4.)
            newattr[tty.CC][termios.VMIN] = 1

            termios.tcsetattr(self.fileno, termios.TCSANOW, newattr)

        self.console.show_cursor(False)
        self.console.file.write("\033[?1003h\n")

        self._key_thread = Thread(
            target=self.run_input_thread, args=(asyncio.get_event_loop(),)
        )
        width, height = self.console.size = self._get_terminal_size()
        asyncio.run_coroutine_threadsafe(
            self._target.post_message(events.Resize(self._target, width, height)),
            loop=loop,
        )
        self._key_thread.start()

    @classmethod
    def _patch_lflag(cls, attrs: int) -> int:
        return attrs & ~(termios.ECHO | termios.ICANON | termios.IEXTEN | termios.ISIG)

    @classmethod
    def _patch_iflag(cls, attrs: int) -> int:
        return attrs & ~(
            # Disable XON/XOFF flow control on output and input.
            # (Don't capture Ctrl-S and Ctrl-Q.)
            # Like executing: "stty -ixon."
            termios.IXON
            | termios.IXOFF
            |
            # Don't translate carriage return into newline on input.
            termios.ICRNL
            | termios.INLCR
            | termios.IGNCR
        )

    def disable_input(self) -> None:
        try:
            if not self.exit_event.is_set():
                signal.signal(signal.SIGWINCH, signal.SIG_DFL)
                self._disable_mouse_support()
                self.exit_event.set()
                if self._key_thread is not None:
                    self._key_thread.join()
        except Exception:
            log.exception("error in disable_input")

    def stop_application_mode(self) -> None:
        log.debug("stop_application_mode()")

        self.disable_input()

        if self.attrs_before is not None:
            try:
                termios.tcsetattr(self.fileno, termios.TCSANOW, self.attrs_before)
            except termios.error:
                pass

        self.console.set_alt_screen(False)
        self.console.show_cursor(True)

    def run_input_thread(self, loop) -> None:
        try:
            self._run_input_thread(loop)
        except Exception:
            log.exception("error running input thread")

    def _run_input_thread(self, loop) -> None:
        def send_event(event: events.Event) -> None:
            asyncio.run_coroutine_threadsafe(
                self._target.post_message(event),
                loop=loop,
            )

        selector = selectors.DefaultSelector()
        selector.register(self.fileno, selectors.EVENT_READ)

        fileno = self.fileno

        def more_data() -> bool:
            """Check if there is more data to parse."""
            for key, events in selector.select(0.1):
                if events:
                    return True
            return False

        parser = XTermParser(self._target, more_data)

        utf8_decoder = getincrementaldecoder("utf-8")().decode
        decode = utf8_decoder
        read = os.read

        mouse_down_time = time()

        log.debug("started key thread")
        try:
            while not self.exit_event.is_set():
                selector_events = selector.select(0.1)
                for _selector_key, mask in selector_events:
                    if mask | selectors.EVENT_READ:
                        unicode_data = decode(read(fileno, 1024))
                        for event in parser.feed(unicode_data):
                            self.process_event(event)
        except Exception:
            log.exception("error running key thread")
        finally:
            selector.close()


if __name__ == "__main__":
    from time import sleep
    from rich.console import Console
    from . import events

    console = Console()

    from .app import App

    class MyApp(App):
        async def on_startup(self, event: events.Startup) -> None:
            self.set_timer(5, callback=self.close_messages)

    MyApp.run()
